@@ -1,6 +1,6 @@
 /*
    Copyright (C) 2004 - 2016 by Guillaume Melquiond <guillaume.melquiond@inria.fr>
-   Part of the Gappa tool http://gappa.gforge.inria.fr/
+   Part of the Gappa tool https://gappa.gitlabpages.inria.fr/
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the CeCILL Free Software License Agreement or
@@ -34,6 +34,7 @@ static char const *theorem_defs[][2] = {
   { "subset_r", "$gpred_bnd.$t $x $1i $u $" },
   { "abs_subset", "$gpred_abs.$t $x $1i $i $" },
   { "rel_subset", "$gpred_rel.$t $x $y $1i $i $" },
+  { "lin_subset", "$gpred_lin.$t $x $y $1i $i $" },
   { "fix_subset", "$gpred_fixflt.$t $x $1c $c $" },
   { "flt_subset", "$gpred_fixflt.$t $x $1c $c $" },
 
@@ -43,7 +44,9 @@ static char const *theorem_defs[][2] = {
   { "intersect_hh", "$gpred_bnd.$t $x $1u $2l $i $" },
   { "intersect_aa", "$gpred_abs.$t $x $1i $2i $i $" },
   { "intersect_rr", "$gpred_rel.$t $x $y $1i $2i $i $" },
+  { "intersect_ll", "$gpred_lin.$t $x $y $1i $2i $i $" },
   { "intersect_rr0", "$gpred_rel.$t $1x $1y $1i $2i $i $" },
+  { "intersect_ll0", "$gpred_lin.$t $1x $1y $1i $2i $i $" },
   { "absurd_intersect", "$gpred_bnd.$t $1x $1i $2i $" },
   { "absurd_intersect_hb", "$gpred_bnd.$t $1x $1u $2i $" },
   { "absurd_intersect_bh", "$gpred_bnd.$t $1x $1i $2l $" },
@@ -100,34 +103,39 @@ static char const *theorem_defs[][2] = {
   { "mul_aa", "$gpred_abs.$t $1x $2x $1i $2i $i $" },
   { "div_aa", "$gpred_abs.$t $1x $2x $1i $2i $i $" },
 
-  { "add_rr", "$gpred_rel.$t $1x $1y $2x $2y $1i $2i $3i $i $1p $2p $3p $4p _" },
-  { "sub_rr", "$gpred_rel.$t $1x $1y $2x $2y $1i $2i $3i $i $1p $2p $3p $4p _" },
+  { "add_rr_left", "$gpred_rel.$t $1x $1y $2x $2y $1i $2i $3i $i $" },
+  { "add_rr_right", "$gpred_rel.$t $1x $1y $2x $2y $1i $2i $3i $i $" },
+  { "sub_rr_left", "$gpred_rel.$t $1x $1y $2x $2y $1i $2i $3i $i $" },
+  { "sub_rr_right", "$gpred_rel.$t $1x $1y $2x $2y $1i $2i $3i $i $" },
   { "mul_rr", "$gpred_rel.$t $1x $1y $2x $2y $1i $2i $i $" },
-  { "div_rr", "$gpred_rel.$t $1x $1y $2x $3x $1i $2i $i $" },
-  { "inv_r", "$gpred_rel.$t $1x $1y _ $1i $i $1p $2p _" },
+  { "div_rr", "$gpred_rel.div_rr' $1x $1y $2x $2y $1i $2i $i $" },
+  { "inv_r", "$gpred_rel.inv_r' $1x $1y _ $1i $i $" },
   { "compose", "$gpred_rel.$t $1x $1y $2y $1i $2i $i $" },
   { "compose_swap", "$gpred_rel.$t $1x $y $2x $3x $1i $2i $i $" },
-  { "bnd_div_of_rel_bnd_div", "$gpred_rel.$t $1x $1y $3x $1i $2i $i $" },
 
   { "abs_fix", "$gpred_fixflt.$t $1x $1c $c $" },
   { "neg_fix", "$gpred_fixflt.$t $1x $1c $c $" },
   { "add_fix", "$gpred_fixflt.$t $1x $2x $1c $2c $c $" },
   { "sub_fix", "$gpred_fixflt.$t $1x $2x $1c $2c $c $" },
   { "mul_fix", "$gpred_fixflt.$t $1x $2x $1c $2c $c $" },
+  { "abs_flt", "$gpred_fixflt.$t $1x $1c $c $" },
+  { "neg_flt", "$gpred_fixflt.$t $1x $1c $c $" },
   { "mul_flt", "$gpred_fixflt.$t $1x $2x $1c $2c $c $" },
-  { "sub_flt", "$gpred_fixflt.$t $1x $2x $1c $2c $3i $c $" },
-  { "sub_flt_rev", "$gpred_fixflt.$t $1x $2x $1c $2c $3i $c $" },
+  { "add_flt_lin", "$gpred_fixflt.$t $1x $2x $1c $2c $3i $c $" },
+  { "sub_flt_lin", "$gpred_fixflt.$t $1x $2x $1c $2c $3i $c $" },
+  { "add_flt_lin_rev", "$gpred_fixflt.$t $1x $2x $1c $2c $3i $c $" },
+  { "sub_flt_lin_rev", "$gpred_fixflt.$t $1x $2x $1c $2c $3i $c $" },
 
   { "neg_nzr", "$gpred_nzr.$t $1x $1p" },
-  { "mul_nzr", "$gpred_nzr.$t $1x $2x $1p $2p" },
-  { "div_nzr", "$gpred_nzr.$t $1x $2x $1p $2p" },
+  { "mul_nzr", "$gpred_nzr.$t $1x $2x $!" },
+  { "div_nzr", "$gpred_nzr.$t $1x $2x $!" },
   { "nzr_of_abs", "$gpred_nzr.$t $x $1i $" },
   { "nzr_of_nzr_rel", "$gpred_nzr.$t $x $1x $2i $" },
-  { "nzr_of_nzr_rel_rev", "$gpred_nzr.$t $1x $x $2i $1p $2p" },
+  { "nzr_of_nzr_rel_rev", "$gpred_nzr.$t $1x $x $2i $!" },
 
   { "rel_refl", "$gpred_rel.$t $x $i $" },
-  { "bnd_of_nzr_rel", "$gpred_rel.$t $2x $1x $i $1p $2p" },
-  { "rel_of_nzr_bnd", "$gpred_rel.$t $x $1x $2i $1p $2p" },
+  { "bnd_of_nzr_rel", "$gpred_rel.$t $2x $1x $i $!" },
+  { "rel_of_nzr_bnd", "$gpred_rel.$t $x $1x $2i $!" },
   { "error_of_rel_pp", "$gpred_rel.$t $1x $2x $1i $2i $i $" },
   { "error_of_rel_po", "$gpred_rel.$t $1x $2x $1i $2i $i $" },
   { "error_of_rel_pn", "$gpred_rel.$t $1x $2x $1i $2i $i $" },
@@ -144,6 +152,40 @@ static char const *theorem_defs[][2] = {
   { "bnd_of_rel_bnd_o", "$gpred_rel.$t $2x $2y $1i $2i $i $" },
   { "bnd_of_rel_bnd_n", "$gpred_rel.$t $2x $2y $1i $2i $i $" },
 
+  { "lin_of_bnd_div", "$gpred_lin.$t $x $y $i $2!" },
+  { "bnd_div_of_lin", "$gpred_lin.$t $2x $2y $2i $!" },
+  { "lin_of_rel", "$gpred_lin.$t $x $y $1i $i $" },
+  { "add_ll", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "sub_ll", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "minus_lin", "$gpred_lin.$t $1x $y $1i $i $" },
+  { "mul_ll_pp", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "mul_ll_po", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "mul_ll_pn", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "mul_ll_op", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "mul_ll_oo", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "mul_ll_on", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "mul_ll_np", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "mul_ll_no", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "mul_ll_nn", "$gpred_lin.$t $1x $2x $1y $2y $1i $2i $i $" },
+  { "compose_lin_pp", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "compose_lin_po", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "compose_lin_pn", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "compose_lin_op", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "compose_lin_oo", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "compose_lin_on", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "compose_lin_np", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "compose_lin_no", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "compose_lin_nn", "$gpred_lin.$t $1x $1y $2y $1i $2i $i $" },
+  { "lin_of_lin_rel_pp", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "lin_of_lin_rel_po", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "lin_of_lin_rel_pn", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "lin_of_lin_rel_op", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "lin_of_lin_rel_oo", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "lin_of_lin_rel_on", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "lin_of_lin_rel_np", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "lin_of_lin_rel_no", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+  { "lin_of_lin_rel_nn", "$gpred_lin.$t $1x $2x $y $1i $2i $i $" },
+
   { "fix_of_singleton_bnd", "$gpred_fixflt.$t $x $1i $c $" },
   { "flt_of_singleton_bnd", "$gpred_fixflt.$t $x $1i $c $" },
   { "fix_of_flt_bnd", "$gpred_fixflt.$t $1x $2i $c $1c $" },
@@ -204,14 +246,16 @@ static char const *theorem_defs[][2] = {
   { "fixed_error_ne", "$gfixed.$t _ _ $i $" },
   { "fixed_error_na", "$gfixed.$t _ _ $i $" },
 
-  { "bnd_rewrite", "$grewriting.$t $1x $1y $i $1p $2p" },
-  { "abs_rewrite", "$grewriting.$t $1x $1y $i $1p $2p" },
-  { "fix_rewrite", "$grewriting.$t $1x $1y $c $1p $2p" },
-  { "flt_rewrite", "$grewriting.$t $1x $1y $c $1p $2p" },
-  { "nzr_rewrite", "$grewriting.$t $1x $1y $1p $2p" },
-  { "rel_rewrite_1", "$grewriting.$t $1x $1y $y $i $1p $2p" },
-  { "rel_rewrite_2", "$grewriting.$t $1x $1y $x $i $1p $2p" },
-  { "eql_trans", "$grewriting.$t $x $1y $y $1p $2p" },
+  { "bnd_rewrite", "$grewriting.$t $1x $1y $i $2!" },
+  { "abs_rewrite", "$grewriting.$t $1x $1y $i $2!" },
+  { "fix_rewrite", "$grewriting.$t $1x $1y $c $2!" },
+  { "flt_rewrite", "$grewriting.$t $1x $1y $c $2!" },
+  { "nzr_rewrite", "$grewriting.$t $1x $1y $2!" },
+  { "rel_rewrite_1", "$grewriting.$t $1x $1y $y $i $2!" },
+  { "rel_rewrite_2", "$grewriting.$t $1x $1y $x $i $2!" },
+  { "lin_rewrite_1", "$grewriting.$t $1x $1y $y $i $2!" },
+  { "lin_rewrite_2", "$grewriting.$t $1x $1y $x $i $2!" },
+  { "eql_trans", "$grewriting.$t $x $1y $y $2!" },
 
   { "opp_mibs", "$grewriting.$t _ _ $i $1p" },
   { "opp_xibu", "$grewriting.$t $x $i $1p" },
@@ -224,52 +268,55 @@ static char const *theorem_defs[][2] = {
   { "add_mibs", "$grewriting.$t _ _ _ _ $i $1p" },
   { "add_fils", "$grewriting.$t _ _ _ $i $1p" },
   { "add_firs", "$grewriting.$t _ _ _ $i $1p" },
-  { "add_filq", "$grewriting.$t _ _ _ $i $1p $2p" },
-  { "add_firq", "$grewriting.$t _ _ _ $i $1p $2p" },
+  { "add_filq", "$grewriting.add_filq' _ _ _ $i $1!" },
+  { "add_firq", "$grewriting.add_firq' _ _ _ $i $1!" },
   { "add_xilu", "$grewriting.$t $x _ $i $1p" },
   { "add_xiru", "$grewriting.$t _ $x $i $1p" },
   { "sub_mibs", "$grewriting.$t _ _ _ _ $i $1p" },
   { "sub_fils", "$grewriting.$t _ _ _ $i $1p" },
   { "sub_firs", "$grewriting.$t _ _ _ $i $1p" },
-  { "sub_filq", "$grewriting.$t _ _ _ $i $1p $2p" },
-  { "sub_firq", "$grewriting.$t _ _ _ $i $1p $2p" },
+  { "sub_filq", "$grewriting.sub_filq' _ _ _ $i $1!" },
+  { "sub_firq", "$grewriting.sub_firq' _ _ _ $i $1!" },
   { "mul_fils", "$grewriting.$t _ _ _ $i $1p" },
   { "mul_firs", "$grewriting.$t _ _ _ $i $1p" },
   { "mul_mals", "$grewriting.$t _ _ _ _ $i $1p" },
   { "mul_mars", "$grewriting.$t _ _ _ _ $i $1p" },
   { "mul_mabs", "$grewriting.$t _ _ _ _ $i $1p" },
   { "mul_mibs", "$grewriting.$t _ _ _ _ $i $1p" },
-  { "mul_xilu", "$grewriting.$t $x _ $i $1p $2p" },
-  { "mul_xiru", "$grewriting.$t _ $x $i $1p $2p" },
-  { "err_xalq", "$grewriting.$t _ $2x $3x $i $1p $2p $3p" },
+  { "mul_xilu", "$grewriting.$t $x _ $i $2!" },
+  { "mul_xiru", "$grewriting.$t _ $x $i $2!" },
+  { "err_xalq", "$grewriting.$t _ $2x $3x $i $!" },
   { "sqrt_mibs", "$grewriting.$t $1x $2x $1i $2i $3i $" },
   { "sqrt_mibq", "$grewriting.$t $1x $2x $1i $2i $3i $" },
   { "sqrt_xibu", "$grewriting.$t $x $1i $2i $" },
   { "val_xebs", "$grewriting.$t $x _ $i $1p" },
   { "val_xabs", "$grewriting.$t _ $x $i $1p" },
-  { "div_mibq", "$grewriting.$t _ $1x $2x $3x $i $1p $2p $3p $4p" },
-  { "div_xals", "$grewriting.$t _ _ $1x $i $1p $2p" },
-  { "div_fil", "$grewriting.$t $1x $2x $i $1p $2p" },
-  { "div_fir", "$grewriting.$t $2x $1x $i $1p $2p" },
-  { "div_firs", "$grewriting.$t _ $1x _ $i $1p $2p" },
-  { "div_xilu", "$grewriting.$t $x _ $i $1p $2p" },
-  { "div_xiru", "$grewriting.$t _ $x $i $1p $2p $3p" },
-  { "err_xabq", "$grewriting.$t _ $1x $i $1p $2p" },
-  { "err_fabq", "$grewriting.$t _ $1x $i $1p $2p" },
-  { "addf_1", "$grewriting.$t $1x _ $i $1p $2p $3p" },
-  { "addf_2", "$grewriting.$t _ _ $i $1p $2p" },
-  { "addf_3", "$grewriting.$t $1x _ $i $1p $2p $3p" },
-  { "addf_4", "$grewriting.$t _ _ $i $1p $2p" },
-  { "addf_5", "$grewriting.$t _ $1x $i $1p $2p $3p" },
-  { "addf_6", "$grewriting.$t _ _ $i $1p $2p" },
-  { "addf_7", "$grewriting.$t _ $1x $i $1p $2p $3p" },
-  { "addf_8", "$grewriting.$t _ _ $i $1p $2p" },
+  { "div_mibq", "$grewriting.$t _ $1x $2x $3x $i $4!" },
+  { "div_xals", "$grewriting.div_xals' _ _ _ $i $1!" },
+  { "div_fil", "$grewriting.$t $1x $2x $i $!" },
+  { "div_fir", "$grewriting.$t $2x $1x $i $!" },
+  { "div_firs", "$grewriting.div_firs' _ _ _ $i $1!" },
+  { "div_xilu", "$grewriting.$t $x _ $i $2!" },
+  { "div_xiru", "$grewriting.div_xiru' $1x $x $i $2!" },
+  { "err_xabq", "$grewriting.$t _ $1x $i $2!" },
+  { "err_fabq", "$grewriting.$t _ $1x $i $2!" },
+  { "addf_9", "$grewriting.$t _ $1x $i $3!" },
+  { "addf_10", "$grewriting.$t $1x _ $i $3!" },
+  { "addf_11", "$grewriting.$t _ $1x $i $3!" },
+  { "addf_12", "$grewriting.$t $1x _ $i $3!" },
 
   { "opp_fibq", "$grewriting.$t $1x $1y $i $1p" },
   { "mul_filq", "$grewriting.$t _ $1x $1y $i $1p" },
   { "mul_firq", "$grewriting.$t $1x _ $1y $i $1p" },
   { "div_firq", "$grewriting.$t $1x _ $1y $i $1p" },
 
+  { "opp_fibm", "$grewriting.$t $1x $1y $i $1p" },
+  { "mul_film", "$grewriting.$t _ $1x $1y $i $1p" },
+  { "mul_firm", "$grewriting.$t $1x _ $1y $i $1p" },
+  { "mul_fism", "$grewriting.$t _ $1x $1y $i $1p" },
+  { "mul_fitm", "$grewriting.$t $1x _ $1y $i $1p" },
+  { "div_firm", "$grewriting.$t $1x _ $1y $i $1p" },
+
   { "opp_fibe", "$grewriting.$t $1x $1y $1p" },
   { "add_file", "$grewriting.$t _ $1x $1y $1p" },
   { "add_fire", "$grewriting.$t $1x _ $1y $1p" },
@@ -295,10 +342,10 @@ RUN_ONCE(fill_theorem_map)
 bool fqn = false;
 bool vernac = true;
 
-static std::string qualify(std::string const &path)
+static char const *qualify(char const *path)
 {
   if (fqn) return path;
-  else return std::string();
+  else return "";
 }
 
 std::string convert_name(std::string const &name)
@@ -409,8 +456,7 @@ std::string display(ast_real const *r)
     return name;
   }
   auto_flush plouf;
-  plouf << (vernac ? "Notation " : "let ") << name << " := ";
-  if (vernac) plouf << '(';
+  plouf << (vernac ? "Definition " : "let ") << name << " := ";
   if (ast_number const *const *nn = boost::get< ast_number const *const >(r))
   {
     ast_number const &n = **nn;
@@ -466,7 +512,7 @@ std::string display(ast_real const *r)
   }
   else
     assert(false);
-  plouf << (vernac ? ").\n" : " in\n");
+  plouf << (vernac ? ".\n" : " in\n");
   return name;
 }
 
@@ -510,6 +556,10 @@ std::string display(property const &p)
         s << qualify(GAPPADEF) << "REL " << display(real) << ' '
           << display(p.real.real2()) << ' ' << display(bnd);
         break;
+      case PRED_LIN:
+        s << qualify(GAPPADEF) << "LIN " << display(real) << ' '
+          << display(p.real.real2()) << ' ' << display(bnd);
+        break;
       default:
         assert(false);
       }
@@ -534,8 +584,8 @@ std::string display(property const &p)
       assert(false);
     }
   }
-  *out << (vernac ? "Notation " : "let ") << name << (vernac ? " := (" : " := ")
-       << s.str() << (vernac ? "). (* " : " in (* ") << dump_property(p) << " *)\n";
+  *out << (vernac ? "Definition " : "let ") << name << " := "
+       << s.str() << (vernac ? ". (* " : " in (* ") << dump_property(p) << " *)\n";
   return name;
 }
 
@@ -559,7 +609,7 @@ std::string display(property_tree const &t)
   std::string name = composite('s', t_id);
   if (t_id < 0) return name;
   auto_flush plouf;
-  plouf << (vernac ? "Definition " : "let ") << name << (vernac ? " := (" : " := ");
+  plouf << (vernac ? "Definition " : "let ") << name << " := ";
   if (!t.left) {
     assert(!t.conjunction);
     plouf << "not " << display(fetch(*t.atom));
@@ -567,10 +617,21 @@ std::string display(property_tree const &t)
     plouf << display(*t.left) << (t.conjunction ? " /\\ " : " \\/ ")
       << display(*t.right);
   }
-  plouf << (vernac ? ").\n" : " in\n");
+  plouf << (vernac ? ".\n" : " in\n");
   return name;
 }
 
+static void apply_hypothesis(auto_flush &plouf, predicated_real const &hr,
+  int hn, property_map const *pmap, int *num)
+{
+  plouf << 'h';
+  if (pmap) {
+    property_map::const_iterator pki = pmap->find(hr);
+    assert(pki != pmap->end());
+    plouf << pki->second.first;
+  } else plouf << (num ? num[hn] : hn);
+}
+
 void apply_theorem(auto_flush &plouf, std::string const &th,
   property const &res, property const *hyp, property_map const *pmap, int *num)
 {
@@ -580,30 +641,30 @@ void apply_theorem(auto_flush &plouf, std::string const &th,
     plouf << "UNKNOWN_" << th;
     return;
   }
-  char const *p = it->second.c_str();
-  int max = 0;
-  std::string buf;
-  for (; *p; ++p)
+  int max = -1;
+  for (char const *p = it->second.c_str(); *p; ++p)
   {
-    if (*p != '$') { plouf << *p; continue; }
-    ++p;
+    char const *q = strchr(p, '$');
+    if (!q) { plouf << p; return; }
+    plouf.write(p, q - p);
+    p = q + 1;
     property const *h = &res;
     if (*p >= '1' && *p <= '9')
     {
       int n = *(p++) - '1';
       h = &hyp[n];
-      for (; max <= n; ++max) {
-        char t[] = { ' ', '$', char('1' + max), 'p', '\0' };
-        buf += t;
-      }
+      if (max < n) max = n;
     }
     switch (*p) {
-      case 'g': {
-        char const *q = strchr(p + 1, '.');
+      case 'g':
+        q = strchr(p + 1, '.');
         assert(q);
-        plouf << qualify("Gappa.Gappa_" + std::string(p + 1, q + 1));
+        if (fqn) {
+          plouf << "Gappa.Gappa_";
+          plouf.write(p + 1, q - p);
+        }
         p = q;
-        break; }
+        break;
       case 't': plouf << th; break;
       case 'i': plouf << display(h->bnd()); break;
       case 'l': plouf << display(lower(h->bnd())); break;
@@ -611,30 +672,29 @@ void apply_theorem(auto_flush &plouf, std::string const &th,
       case 'c': plouf << '(' << h->cst() << ')'; break;
       case 'x': plouf << display(h->real.real()); break;
       case 'y': plouf << display(h->real.real2()); break;
-      case '\0':
-        buf += " _";
-        p = buf.c_str();
-        // the first character will be skipped
-        assert (*p == ' ');
-        break;
       case 'p':
-        plouf << 'h';
-        if (pmap) {
-          property_map::const_iterator pki = pmap->find(h->real);
-          assert(pki != pmap->end());
-          plouf << pki->second.first;
-        } else if (num)
-          plouf << num[h - hyp];
-        else
-          plouf << h - hyp;
+        apply_hypothesis(plouf, h->real, h - hyp, pmap, num);
         break;
+      case '!':
+        for (int n = 0; n <= max; ++n) {
+          apply_hypothesis(plouf, hyp[n].real, n, pmap, num);
+          if (n != max) plouf << ' ';
+        }
+        break;
+      case '\0':
+        for (int n = 0; n <= max; ++n) {
+          apply_hypothesis(plouf, hyp[n].real, n, pmap, num);
+          plouf << ' ';
+        }
+        plouf << '_';
+        return;
       default:
         plouf << '$';
     }
   }
 }
 
-std::string display(theorem_node *t)
+std::string display(theorem_node const *t)
 {
   static int t_id = 0;
   std::string name = composite('t', ++t_id);
@@ -642,9 +702,7 @@ std::string display(theorem_node *t)
   if (vernac)
   {
     plouf << "Lemma " << name << " : ";
-    for (property_vect::const_iterator i = t->hyp.begin(),
-         end = t->hyp.end(); i != end; ++i)
-      plouf << display(*i) << " -> ";
+    for (property const &i: t->hyp) plouf << display(i) << " -> ";
     plouf << display(t->res) << ".\nProof.\n";
     int nb_hyps = t->hyp.size();
     if (nb_hyps)
@@ -659,9 +717,9 @@ std::string display(theorem_node *t)
   {
     plouf << "let " << name;
     int num_hyp = 0;
-    for (property_vect::const_iterator i = t->hyp.begin(),
-         i_end = t->hyp.end(); i != i_end; ++i)
-      plouf << " (h" << num_hyp++ << " : " << display(*i) << ')';
+    for (property const &i: t->hyp) {
+      plouf << " (h" << num_hyp++ << " : " << display(i) << ')';
+    }
     plouf << " : " <<  display(t->res) << " := ";
   }
 
@@ -684,6 +742,7 @@ std::string subset_name(property const &p1, property const &p2)
     break;
   case PRED_ABS: prefix = "abs_"; break;
   case PRED_REL: prefix = "rel_"; break;
+  case PRED_LIN: prefix = "lin_"; break;
   case PRED_FIX: prefix = "fix_"; break;
   case PRED_FLT: prefix = "flt_"; break;
   case PRED_EQL:
@@ -695,19 +754,18 @@ std::string subset_name(property const &p1, property const &p2)
 
 void invoke_lemma(auto_flush &plouf, property_vect const &hyp, property_map const &pmap)
 {
-  for (property_vect::const_iterator j = hyp.begin(),
-       j_end = hyp.end(); j != j_end; ++j)
+  for (property const &j: hyp)
   {
-    property_map::const_iterator pki = pmap.find(j->real);
+    property_map::const_iterator pki = pmap.find(j.real);
     assert(pki != pmap.end());
     int h = pki->second.first;
-    std::string sn = subset_name(*pki->second.second, *j);
+    std::string sn = subset_name(*pki->second.second, j);
     if (sn.empty()) {
       if (vernac) plouf << " exact h" << h << '.';
       else plouf << " h" << h;
     } else {
       plouf << (vernac ? " refine (" : " (");
-      apply_theorem(plouf, sn, *j, pki->second.second, &pmap);
+      apply_theorem(plouf, sn, j, pki->second.second, &pmap);
       plouf << (vernac ? ") ; finalize." : ")");
     }
   }
@@ -717,13 +775,13 @@ void invoke_lemma(auto_flush &plouf, property_vect const &hyp, property_map cons
 /**
  * Instantiate node @a m as "h @a num_hyp" with the hypotheses of node @a n.
  */
-static void pose_hypothesis(auto_flush &plouf, int num_hyp, node *m, node *n)
+static void pose_hypothesis(auto_flush &plouf, int num_hyp, node const *m, node const *n)
 {
   plouf << (vernac ? " assert (h" : "  let h") << num_hyp << " := ";
   int i = 0;
   graph_t *g = n->graph;
   for (; g != m->graph; g = g->get_father()) ++i;
-  if (m->type == LOGIC && !static_cast<logic_node *>(m)->before) {
+  if (m->type == LOGIC && !static_cast<logic_node const *>(m)->before) {
     plouf << 'h' << i;
   } else {
     plouf << display(m);
@@ -765,6 +823,9 @@ static void reify(auto_flush &plouf, real_map &rm, property const &p)
   case PRED_REL:
     plouf << "Arel " << find_real(rm, p.real.real()) << "%nat " << find_real(rm, p.real.real2()) << "%nat " << display(p.bnd());
     break;
+  case PRED_LIN:
+    plouf << "Alin " << find_real(rm, p.real.real()) << "%nat " << find_real(rm, p.real.real2()) << "%nat " << display(p.bnd());
+    break;
   case PRED_FIX:
     plouf << "Afix " << find_real(rm, p.real.real()) << "%nat (" << p.cst() << ")%Z";
     break;
@@ -812,14 +873,11 @@ static void simplification(auto_flush &plouf, property_tree const &before, prope
   reify(plouf, rm, mod);
   plouf << ") ";
   std::vector<ast_real const *> inv(rm.size(), NULL);
-  for (real_map::const_iterator i = rm.begin(), i_end = rm.end(); i != i_end; ++i)
-  {
-    inv[i->second] = i->first;
+  for (auto const &i: rm) {
+    inv[i.second] = i.first;
   }
-  for (std::vector<ast_real const *>::const_iterator i = inv.begin(),
-       i_end = inv.end(); i != i_end; ++i)
-  {
-    plouf << "(List.cons " << display(*i) << ' ';
+  for (ast_real const *i: inv) {
+    plouf << "(List.cons " << display(i) << ' ';
   }
   plouf << "List.nil" << std::string(rm.size(), ')')
      << " h" << num_hyp << " h" << (num_hyp - 1) << " _";
@@ -834,9 +892,9 @@ static void select(auto_flush &plouf, int idx, int num_hyp)
   if (vernac) plouf << ").\n";
 }
 
-static id_cache< node * > displayed_nodes;
+static id_cache<node const *> displayed_nodes;
 
-std::string display(node *n)
+std::string display(node const *n)
 {
   assert(n);
   int n_id = displayed_nodes.find(n);
@@ -851,9 +909,9 @@ std::string display(node *n)
     plouf << display(g->get_hypotheses()) << " -> ";
     ++num_hyp;
   }
-  logic_node *ln = NULL;
+  logic_node const *ln = NULL;
   if (n->type == LOGIC) {
-    ln = static_cast<logic_node *>(n);
+    ln = static_cast<logic_node const *>(n);
     assert(ln->before);
     plouf << display(ln->tree);
   } else {
@@ -881,7 +939,7 @@ std::string display(node *n)
     }
     break; }
   case LOGICP: {
-    logicp_node *ln = static_cast<logicp_node *>(n);
+    logicp_node const *ln = static_cast<logicp_node const *>(n);
     assert(ln->before && ln->before->tree.conjunction);
     pose_hypothesis(plouf, num_hyp, ln->before, n);
     select(plouf, ln->index, num_hyp);
@@ -889,15 +947,12 @@ std::string display(node *n)
   case MODUS: {
     property_map pmap;
     node_vect const &pred = n->get_subproofs();
-    for (node_vect::const_iterator i = pred.begin(),
-         i_end = pred.end(); i != i_end; ++i)
-    {
-      node *m = *i;
+    for (node const *m: pred) {
       pose_hypothesis(plouf, num_hyp, m, n);
       property const &res = m->get_result();
       pmap[res.real] = std::make_pair(num_hyp++, &res);
     }
-    modus_node *mn = static_cast<modus_node *>(n);
+    modus_node const *mn = static_cast<modus_node const *>(n);
     assert(mn->target);
     plouf << (vernac ? " apply " : "  ") << display(mn->target);
     if (vernac) plouf << '.';
@@ -928,6 +983,9 @@ std::string display(node *n)
         case PRED_REL:
           suffix = res.real == n_res.real ? "_rr" : "_rr0";
           break;
+        case PRED_LIN:
+          suffix = res.real == n_res.real ? "_ll" : "_ll0";
+          break;
         default:
           assert(false);
       }
@@ -948,10 +1006,8 @@ std::string display(node *n)
     node_vect const &pred = n->get_subproofs();
     property const &n_res = n->get_result();
     assert(pred.size() >= 2);
-    for (node_vect::const_iterator i = pred.begin(),
-         i_end = pred.end(); i != i_end; ++i)
+    for (node const *m: pred)
     {
-      node *m = *i;
       assert(m->graph->father == n->graph);
       property const &p = *m->graph->get_hypotheses().atom;
       if (upper(p.bnd()) < number::pos_inf) {
