@@ -1,6 +1,6 @@
 /*
-   Copyright (C) 2004 - 2015 by Guillaume Melquiond <guillaume.melquiond@inria.fr>
-   Part of the Gappa tool http://gappa.gforge.inria.fr/
+   Copyright (C) 2004 - 2025 by Guillaume Melquiond <guillaume.melquiond@inria.fr>
+   Part of the Gappa tool https://gappa.gitlabpages.inria.fr/
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the CeCILL Free Software License Agreement or
@@ -51,9 +51,9 @@ preal_vect rewriting_scheme::needed_reals(predicated_real const &d,
   pattern_cond_vect const &c) const
 {
   preal_vect res;
-  for (pattern_cond_vect::const_iterator i = c.begin(), i_end = c.end();
-       i != i_end; ++i)
-    res.push_back(predicated_real(i->real, i->type == COND_NZ ? PRED_NZR : PRED_BND));
+  for (pattern_cond const &i: c) {
+    res.push_back(predicated_real(i.real, i.type == COND_NZ ? PRED_NZR : PRED_BND));
+  }
   if (!d.null()) res.push_back(d);
   return res;
 }
@@ -100,16 +100,15 @@ static rewriting_scheme *generate_rewriting_scheme(predicated_real const &src,
 {
   if (src.real()->is_constant && !src.real2()) return NULL;
 
-  for (pattern_excl_vect::const_iterator i = rule->excl.begin(),
-       i_end = rule->excl.end(); i != i_end; ++i)
-  {
-    if (rewrite(i->first, holders) == rewrite(i->second, holders))
+  for (pattern_excl const &i: rule->excl) {
+    if (rewrite(i.first, holders) == rewrite(i.second, holders))
       return NULL;
   }
 
   pattern_cond_vect c(rule->cond);
-  for (pattern_cond_vect::iterator i = c.begin(), end = c.end(); i != end; ++i)
-    i->real = rewrite(i->real, holders);
+  for (pattern_cond &i: c) {
+    i.real = rewrite(i.real, holders);
+  }
   return new rewriting_scheme(src, dst, name.c_str(), c);
 }
 
@@ -184,10 +183,8 @@ proof_scheme *proxy_rewriting_factory::operator()(predicated_real const &src,
 
   if (excl)
   {
-    for (pattern_excl_vect::const_iterator i = excl->begin(),
-         i_end = excl->end(); i != i_end; ++i)
-    {
-      if (rewrite(i->first, holders) == rewrite(i->second, holders))
+    for (pattern_excl const &i: *excl) {
+      if (rewrite(i.first, holders) == rewrite(i.second, holders))
         return NULL;
     }
   }
@@ -196,9 +193,7 @@ proof_scheme *proxy_rewriting_factory::operator()(predicated_real const &src,
   ast_real const *r = rewrite(dst.real(), holders);
   preal_vect needed;
   needed.push_back(predicated_real(rewrite(rsrc, holders), rewrite(unhide(rdst), holders), PRED_EQL));
-  needed.push_back(t == PRED_REL
-    ? predicated_real(r, rewrite(dst.real2(), holders), t)
-    : predicated_real(r, t));
+  needed.push_back(predicated_real(r, dst.real2() ? rewrite(dst.real2(), holders) : NULL, t));
   return new proxy_rewriting_scheme(src, needed, name.c_str());
 }
 
@@ -221,6 +216,10 @@ static void generate_all_proxy_rewrite(ast_real const *src,
     pr(dst, free, PRED_REL), "rel_rewrite_1", e);
   new proxy_rewriting_factory(src, dst, pr(free, src, PRED_REL),
     pr(free, dst, PRED_REL), "rel_rewrite_2", e);
+  new proxy_rewriting_factory(src, dst, pr(src, free, PRED_LIN),
+    pr(dst, free, PRED_LIN), "lin_rewrite_1", e);
+  new proxy_rewriting_factory(src, dst, pr(free, src, PRED_LIN),
+    pr(free, dst, PRED_LIN), "lin_rewrite_2", e);
 }
 
 RUN_ONCE(approximate_rewrite) {
@@ -232,10 +231,8 @@ void register_user_rewrite(ast_real const *r1, ast_real const *r2, hint_cond_vec
   pattern_cond_vect pc;
   if (hc)
   {
-    for (hint_cond_vect::const_iterator i = hc->begin(),
-         i_end = hc->end(); i != i_end; ++i)
+    for (hint_cond const *c: *hc)
     {
-      hint_cond const *c = *i;
       pattern_cond cond;
       cond.real = c->real;
       cond.value = atoi(c->value->mantissa.c_str());
@@ -287,10 +284,11 @@ static pattern a(0), b(1), c(2), d(3), a_b(-1), one(token_one);
 
 /*
 Naming convention: operator name followed by
-  x (expand term)      f (factor term)      m (mix)
-  a (approximate expr) e (accurate expr)    i (irrelevant)
-  l (left hand side)   r (right hand side)  b (both)
-  s (absolute error)   q (relative error)   e (equality)   u (user term)
+- x (expand term), f (factor term), m (mix)
+- a (approximate expr), e (accurate expr), i (irrelevant)
+- l (left hand side), r (right hand side), b (both), s t (skewed)
+- s (absolute error), q (relative error), m (relative magnitude),
+  e (equality), u (user term)
 */
 
 #define REWRITE(name,p1,p2) \
@@ -355,16 +353,14 @@ REWRITe(add_firs,
 	a - c,
 	a ^ c);
 
-REWRIT9(add_filq,
+REWRITe(add_filq,
 	((a + b) - (a + c)) / (a + c),
 	(b - c) / (a + c),
-	~(a + c),
 	b ^ c);
 
-REWRIT9(add_firq,
+REWRITe(add_firq,
 	((a + b) - (c + b)) / (c + b),
 	(a - c) / (c + b),
-	~(c + b),
 	a ^ c);
 
 R3WRITE(add_xilu,
@@ -404,16 +400,14 @@ REWRITe(sub_firs,
 	a - c,
 	a ^ c);
 
-REWRIT9(sub_filq,
+REWRITe(sub_filq,
 	((a - b) - (a - c)) / (a - c),
 	hide(-((b - c) / (a - c))),
-	~(a - c),
 	b ^ c);
 
-REWRIT9(sub_firq,
+REWRITe(sub_firq,
 	((a - b) - (c - b)) / (c - b),
 	(a - c) / (c - b),
-	~(c - b),
 	a ^ c);
 
 R3WRITE(val_xabs, // actually sub_xilu
@@ -480,10 +474,9 @@ R3WRIT3(mul_xiru,
 
 // DIV
 
-REWRIT9(div_xals,
+REWRITe(div_xals,
 	b / c,
 	hide((b - a) / c + a / c),
-	~c,
 	a ^ c && b ^ c);
 
 REWRIT3(div_fir,
@@ -496,11 +489,12 @@ REWRIT3(div_fil,
 	b,
 	~a);
 
-REWRIT9(div_firs,
+/*
+REWRITe(div_firs,
 	a / b - c / b,
 	hide((a - c) / b),
-	~b,
 	a ^ c);
+*/
 
 R3WRIT3(div_xilu,
 	a,
@@ -512,7 +506,7 @@ R3WRIT3(div_xiru,
 	b,
 	hide(a / (a / b)),
 	a / b,
-	~a && ~b);
+	~a);
 
 // SQRT
 
@@ -574,74 +568,35 @@ REWRIT3(square_sqrt,
 	a,
 	a >= 0);
 
-REWRIT3(addf_1,
+REWRIT3(addf_9,
 	a / (a + b),
-	hide(one / hide(one + b / a)),
+	hide(one - hide(one / hide(one + a / b))),
+	~b && ~(a + b));
+
+REWRIT3(addf_10,
+	b / (a + b),
+	hide(one - hide(one / hide(one + b / a))),
 	~a && ~(a + b));
 
-REWRIT9(addf_2,
-	a / (a + b),
-	hide(one - b / (a + b)),
-	~(a + b),
-	a ^ b);
-
-REWRIT9(addf_3,
+REWRIT3(addf_11,
 	a / (a - b),
-	hide(one / hide(one - b / a)),
-	~a && ~(a - b),
-	a ^ b);
+	hide(one - hide(one / hide(one - a / b))),
+	~b && ~(a - b));
 
-REWRIT9(addf_4,
-	a / (a - b),
-	hide(one + b / (a - b)),
-	~(a - b),
-	a ^ b);
-
-REWRIT9(addf_5,
-	b / (a + b),
-	hide(one / hide(a / b + one)),
-	~b && ~(a + b),
-	a ^ b);
-
-REWRIT9(addf_6,
-	b / (a + b),
-	hide(one - a / (a + b)),
-	~(a + b),
-	a ^ b);
-
-REWRIT9(addf_7,
+REWRIT3(addf_12,
 	b / (a - b),
-	hide(one / hide(a / b - one)),
-	~b && ~(a - b),
-	a ^ b);
-
-REWRIT9(addf_8,
-	b / (a - b),
-	hide(a / (a - b) - one),
-	~(a - b),
-	a ^ b);
+	hide(hide(one / hide(one - b / a)) - one),
+	~a && ~(a - b));
 
 #undef REWRITE
 #undef REWRIT3
 #undef REWRITe
 #undef REWRIT9
 
-#define REWRITE(name,lhs1,rhs1,lhs2,rhs2) \
-  static rewriting_rule REWRITING_NAME \
-    (predicated_real(lhs1, rhs1, PRED_REL), predicated_real(lhs2, rhs2, PRED_REL),\
-     #name, pattern_cond_vect(), pattern_excl_vect(), NULL)
-#define REWRIT3(name,lhs1,rhs1,lhs2,rhs2,cond) \
-  static rewriting_rule REWRITING_NAME \
-    (predicated_real(lhs1, rhs1, PRED_REL), predicated_real(lhs2, rhs2, PRED_REL),\
-     #name, pattern_cond_vect() && cond, pattern_excl_vect(), NULL)
 #define REWRITe(name,lhs1,rhs1,lhs2,rhs2,excl) \
   static rewriting_rule REWRITING_NAME \
     (predicated_real(lhs1, rhs1, PRED_REL), predicated_real(lhs2, rhs2, PRED_REL),\
      #name, pattern_cond_vect(), pattern_excl_vect() && excl, NULL)
-#define REWRIT9(name,lhs1,rhs1,lhs2,rhs2,cond,excl) \
-  static rewriting_rule REWRITING_NAME \
-    (predicated_real(lhs1, rhs1, PRED_REL), predicated_real(lhs2, rhs2, PRED_REL),\
-     #name, pattern_cond_vect() && cond, pattern_excl_vect() && excl, NULL)
 
 REWRITe(opp_fibq,
 	-a, -b,
@@ -663,27 +618,52 @@ REWRITe(div_firq,
 	a, c,
 	a ^ c);
 
-#undef REWRITE
-#undef REWRIT3
 #undef REWRITe
-#undef REWRIT9
 
 #define REWRITE(name,lhs1,rhs1,lhs2,rhs2) \
   static rewriting_rule REWRITING_NAME \
-    (predicated_real(lhs1, rhs1, PRED_EQL), predicated_real(lhs2, rhs2, PRED_EQL),\
+    (predicated_real(lhs1, rhs1, PRED_LIN), predicated_real(lhs2, rhs2, PRED_LIN),\
      #name, pattern_cond_vect(), pattern_excl_vect(), NULL)
-#define REWRIT3(name,lhs1,rhs1,lhs2,rhs2,cond) \
+#define REWRITe(name,lhs1,rhs1,lhs2,rhs2,excl) \
   static rewriting_rule REWRITING_NAME \
-    (predicated_real(lhs1, rhs1, PRED_EQL), predicated_real(lhs2, rhs2, PRED_EQL),\
-     #name, pattern_cond_vect() && cond, pattern_excl_vect(), NULL)
+    (predicated_real(lhs1, rhs1, PRED_LIN), predicated_real(lhs2, rhs2, PRED_LIN),\
+     #name, pattern_cond_vect(), pattern_excl_vect() && excl, NULL)
+
+REWRITe(opp_fibm,
+	-a, -b,
+	a, b,
+	a ^ b);
+
+REWRITe(mul_film,
+	a * b, a * c,
+	b, c,
+	b ^ c);
+
+REWRITe(mul_firm,
+	a * b, c * a,
+	b, c,
+	a ^ c);
+
+REWRITE(mul_fism,
+	a * b, a * c,
+	b, c);
+
+REWRITE(mul_fitm,
+	a * b, b * c,
+	a, c);
+
+REWRITe(div_firm,
+	a / b, c / b,
+	a, c,
+	a ^ c);
+
+#undef REWRITE
+#undef REWRITe
+
 #define REWRITe(name,lhs1,rhs1,lhs2,rhs2,excl) \
   static rewriting_rule REWRITING_NAME \
     (predicated_real(lhs1, rhs1, PRED_EQL), predicated_real(lhs2, rhs2, PRED_EQL),\
      #name, pattern_cond_vect(), pattern_excl_vect() && excl, NULL)
-#define REWRIT9(name,lhs1,rhs1,lhs2,rhs2,cond,excl) \
-  static rewriting_rule REWRITING_NAME \
-    (predicated_real(lhs1, rhs1, PRED_EQL), predicated_real(lhs2, rhs2, PRED_EQL),\
-     #name, pattern_cond_vect() && cond, pattern_excl_vect() && excl, NULL)
 
 REWRITe(opp_fibe,
 	-a, -b,
@@ -729,3 +709,5 @@ REWRITe(div_fire,
 	a / b, c / b,
 	a, c,
 	a ^ c);
+
+#undef REWRITe
